/**
 ****************************************************************************************
 *
 * @file ble_peripheral_task.c
 *
 * @brief BLE peripheral task
 *
 * Copyright (C) 2015 Dialog Semiconductor.
 * This computer program includes Confidential, Proprietary Information
 * of Dialog Semiconductor. All Rights Reserved.
 *
 ****************************************************************************************
 */

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "osal.h"
#include "time.h"
#include "sys_watchdog.h"
#include "ble_att.h"
#include "ble_common.h"
#include "ble_gap.h"
#include "ble_gatts.h"
#include "ble_service.h"
#include "ble_uuid.h"


/* Required header files */
#include "ad_nvms.h"
#include "ble_custom_service_mechanism.h"


/*
 * Macro for changing the default Maximum Protocol Unit (MTU).
 * The default value is 65 octets (secure connections enabled).
 *
 * 0: Do not change the current MTU
 * 1: Change the current MTU
 */
#define MCS_CHANGE_MTU_SIZE_EN  (0)



/* Maximum allowable Characteristic Attribute value length */
#define MAX_CHAR_ATTR_VAL       (50)


/*
 * Macro for enabling debugging messages on the serial console. Valid values:
 *
 * 1: Enable debugging messages on the serial console
 * 0: Disable debugging messages on the serial console
 */
#define DBG_SERIAL_CONSOLE      (1)


/*
 * Characteristic Attribute value.
 */
__RETAINED_RW uint8_t characteristic_value[MAX_CHAR_ATTR_VAL] = { 0 };



/*
 * Notification bits reservation
 * bit #0 is always assigned to BLE event queue notification
 */


/*
 * BLE peripheral advertising data
 */
static const uint8_t adv_data[] = {
        0x12, GAP_DATA_TYPE_LOCAL_NAME,
        'M', 'y', ' ', 'C', 'u', 's', 't', 'o', 'm', ' ', 'S', 'e', 'r', 'v', 'i', 'c', 'e'
};


/* Task handle */
__RETAINED_RW static OS_TASK ble_task_handle = NULL;



/*
 * This callback will hit upon a read request from a peer device. This implies
 * that the peer device wants to read the Characteristic Attribute value. User
 * is responsible for providing the requested data.
 *
 * \param [in] value: The value that should be returned back to the peer device
 *
 * \param [in] length: The number of bytes/octets of the returned value
 *
 * \note: The callback function should have that specific prototype
 */
void get_var_value_cb(uint8_t **value, uint16_t *length)
{

        /* Return the requested data to the peer device! */
        *value  = characteristic_value;       // A pointer that points to the returned data
        *length = sizeof(MAX_CHAR_ATTR_VAL);  // The size of the returned data, expressed in bytes.

        /*
         * This is just for debugging/demonstration purposes. UART is a slow interface
         * and will add significant delay compared to the BLE speeds. Please note that
         * the BLE stack will not proceed with the next BLE event until the callback
         * returns!
         */
#if (DBG_SERIAL_CONSOLE == 1)
        printf("\nRead callback function hit! - Returned value: %s\n\r", (char *)characteristic_value);
#endif
}


/*
 *
 * @ Write request callback
 *
 * This callback will hit upon a write request from a peer device. This implies
 * that the peer device requested to modify the Characteristic Attribute value.
 *
 * \param [out] value:  The new value for the Characteristic Attribute, written by the peer device.
 *
 * \param [out] length: The length of the new written value, expressed in bytes/octets.
 *
 * \note: The callback function should have that specific prototype
 *
 */
void set_var_value_cb(const uint8_t *value, uint16_t length)
{
        /* Clear the current Characteristic Attribute value */
        memset((void *)characteristic_value, 0x20, sizeof(characteristic_value));

        /* Update the Characteristic Attribute value as requested by the peer device */
        memcpy((void *)characteristic_value, (void *)value, length);

        /*
         * This is just for debugging/demonstration purposes. UART is a slow interface
         * and will add significant delay compared to the BLE speeds. Please note that
         * the BLE stack will not proceed with the next BLE event until the callback
         * returns
         */
#if (DBG_SERIAL_CONSOLE == 1)
        printf("\nWrite callback function hit! - Written value: %s, length: %d\n\r",
                                                        characteristic_value, length);
#endif
}


/*
 * @ Notification event callback
 *
 *  An notification callback will hit for each peer connected device, given that
 *  the peer device has its notifications/indications enabled.
 *
 * \param [in] conn_idx: The connection index. This parameter should be used to identify for which peer device
 *                       a notification/indication has been sent to (useful when more than one device is connected
 *                       to the target peripheral).
 *
 * \param [in] status: This parameter signifies the status of the aforementioned operation:
 *                     0 -> Signifies that a notification/indication was not sent successfully
 *                     1 -> Signifies that a notification/indication was sent successfully
 *
 * \param [in] type: Signifies whether a notification or indication has been sent to the peer device:
 *                   0 -> Notifications,  1-> Indications
 *
 */
void event_sent_cb(uint16_t conn_idx, bool status, gatt_event_t type)
{
        /*
         * This is just for debugging/demonstration purposes. UART is a slow interface
         * and will add significant delay compared to the BLE speeds. Please note that
         * the BLE stack will not proceed with the next BLE event until the callback
         * returns.
         */
#if (DBG_SERIAL_CONSOLE == 1)
        printf("\nNotify callback - Connection idx: %d, Status: %d, Type: %d\n\r",
                                                                conn_idx, status, type);
#endif
}



/*
 * Main code
 */
static void handle_evt_gap_connected(ble_evt_gap_connected_t *evt)
{
        /*
         * Manage connection information
         */
}

static void handle_evt_gap_disconnected(ble_evt_gap_disconnected_t *evt)
{
        /*
         * Manage disconnection information
         */
}

static void handle_evt_gap_adv_completed(ble_evt_gap_adv_completed_t *evt)
{
        // restart advertising so we can connect again
        ble_gap_adv_start(GAP_CONN_MODE_UNDIRECTED);
}



void ble_peripheral_task(void *params)
{
        int8_t wdog_id;
        ble_service_t *svc;

        // in case services which do not use svc are all disabled, just suppress -Wunused-variable
        (void) svc;

        /* register ble_peripheral task to be monitored by watchdog */
        wdog_id = sys_watchdog_register(false);

        /* Get task's handler */
        ble_task_handle = OS_GET_CURRENT_TASK();

        srand(time(NULL));

        ble_peripheral_start();
        ble_register_app();

        ble_gap_device_name_set("Custom BLE Service", ATT_PERM_READ);


#if (MCS_CHANGE_MTU_SIZE_EN == 1)
        uint16_t mtu_size;
        ble_error_t mtu_err;

        /*
         * Get the old MTU size and print it on the serial console.
         */
        mtu_err = ble_gap_mtu_size_get(&mtu_size);
        printf("Old MTU size: %d, Status: %d\n\r", mtu_size, mtu_err);

        /*
         * @brief Change the MTU size.
         *
         * \note: The maximum supported MTU size is 512 octets.  The minimum supported MTU size, as defined
         *        by Bluetooth SIG, is 65 octets when LE secure connections are used, 23 otherwise.
         *
         * \warning: MTU size change should take place prior to creating the BLE attribute database.
         *           Otherwise, any already defined attribute database will be deleted!!!
         */
        mtu_err = ble_gap_mtu_size_set(125);

        /*
         * Get the updated MTU size and print it on the serial console.
         */
        mtu_err = ble_gap_mtu_size_get(&mtu_size);
        printf("New MTU size: %d, Status: %d\n\r", mtu_size, mtu_err);
#endif


        //************ Characteristic declarations for the initialized BLE Service  *************
        const mcs_characteristic_config_t my_service_1[] = {

                CHARACTERISTIC_DECLARATION(11111111-0000-0000-0000-000000000001, MAX_CHAR_ATTR_VAL,
                        CHAR_WRITE_PROP_EN, CHAR_READ_PROP_EN, CHAR_NOTIF_INDIC_EN, Initialized Characteristic,
                                                        get_var_value_cb, set_var_value_cb, event_sent_cb),


                 // -----------------------------------------------------------------
                 // -- Here you can continue adding more Characteristic Attributes --
                 // -----------------------------------------------------------------


        };
        // ***************** Register the Bluetooth Service in Dialog BLE framework *****************
        SERVICE_DECLARATION(my_service_1, 11111111-0000-0000-0000-111111111111)



        //************ Characteristic declarations for the uninitialized BLE Service *************
        const mcs_characteristic_config_t my_service_2[] = {

                /* Uninitialized Characteristic Attribute. You can defined your preferred settings */
                CHARACTERISTIC_DECLARATION(22222222-0000-0000-0000-000000000001, 0,
                          CHAR_WRITE_PROP_DIS, CHAR_READ_PROP_EN, CHAR_NOTIF_NONE, NULL,
                                                                                      NULL, NULL,NULL),


                /* Uninitialized Characteristic Attribute. You can defined your preferred settings */
                CHARACTERISTIC_DECLARATION(22222222-0000-0000-0000-000000000002, 0,
                          CHAR_WRITE_PROP_DIS, CHAR_READ_PROP_DIS, CHAR_NOTIF_NONE, NULL,
                                                                                     NULL, NULL, NULL),


               /* Uninitialized Characteristic Attribute. You can defined your preferred settings */
               CHARACTERISTIC_DECLARATION(22222222-0000-0000-0000-000000000003, 0,
                          CHAR_WRITE_PROP_DIS, CHAR_READ_PROP_DIS, CHAR_NOTIF_NONE, NULL,
                                                                                    NULL, NULL, NULL),


                // -----------------------------------------------------------------
                // -- Here you can continue adding more Characteristic attributes --
                // -----------------------------------------------------------------

       };
       // ****************** Register the Bluetooth Service in Dialog BLE framework *****************
       SERVICE_DECLARATION(my_service_2, 22222222-0000-0000-0000-222222222222)



        ble_gap_adv_data_set(sizeof(adv_data), adv_data, 0, NULL);
        ble_gap_adv_start(GAP_CONN_MODE_UNDIRECTED);



        for (;;) {
                OS_BASE_TYPE ret;
                uint32_t notif;

                /* notify watchdog on each loop */
                sys_watchdog_notify(wdog_id);

                /* suspend watchdog while blocking on OS_TASK_NOTIFY_WAIT() */
                sys_watchdog_suspend(wdog_id);

                /*
                 * Wait on any of the notification bits, then clear them all
                 */
                ret = OS_TASK_NOTIFY_WAIT(0, OS_TASK_NOTIFY_ALL_BITS, &notif, OS_TASK_NOTIFY_FOREVER);
                /* Blocks forever waiting for task notification. The return value must be OS_OK */
                OS_ASSERT(ret == OS_OK);

                /* resume watchdog */
                sys_watchdog_notify_and_resume(wdog_id);

                /* notified from BLE manager, can get event */
                if (notif & BLE_APP_NOTIFY_MASK) {
                        ble_evt_hdr_t *hdr;

                        hdr = ble_get_event(false);
                        if (!hdr) {
                                goto no_event;
                        }

                        if (ble_service_handle_event(hdr)) {
                                goto handled;
                        }

                        switch (hdr->evt_code) {
                        case BLE_EVT_GAP_CONNECTED:
                                handle_evt_gap_connected((ble_evt_gap_connected_t *) hdr);
                                break;
                        case BLE_EVT_GAP_ADV_COMPLETED:
                                handle_evt_gap_adv_completed((ble_evt_gap_adv_completed_t *) hdr);
                                break;
                        case BLE_EVT_GAP_DISCONNECTED:
                                handle_evt_gap_disconnected((ble_evt_gap_disconnected_t *) hdr);
                                break;
                        case BLE_EVT_GAP_PAIR_REQ:
                        {
                                ble_evt_gap_pair_req_t *evt = (ble_evt_gap_pair_req_t *) hdr;
                                ble_gap_pair_reply(evt->conn_idx, true, evt->bond);
                                break;
                        }
                        default:
                                ble_handle_event_default(hdr);
                                break;
                        }

handled:
                        OS_FREE(hdr);

no_event:
                        // notify again if there are more events to process in queue
                        if (ble_has_event()) {
                                OS_TASK_NOTIFY(OS_GET_CURRENT_TASK(), BLE_APP_NOTIFY_MASK, eSetBits);
                        }

                }

        }
}
